from ._helpers import *

import unittest
import numpy as np
import os

class TestBlobInterpreter(unittest.TestCase):
    def setUp(self):
        # Make database
        self.d = sew.Database(":memory:")

        self.fmt = sew.FormatSpecifier(
            [
                ["data", "BLOB"]
            ]
        )
        self.tablename = 'table1'
        self.d.createTable(self.fmt.generate(), self.tablename)
        self.d.reloadTables()

        # Insert some test data
        self.data = {
            'p1': np.array([3], np.uint8),
            'p2': np.array([123], np.int64),
            'p3': np.array([1142.2], np.float64)
        }

        toInsert = bytes()
        for k, v in self.data.items():
            toInsert += v.tobytes()
        self.d[self.tablename].insertOne(toInsert, commitNow=True)

    #%%
    def test_interpret_simple(self):
        # Define interpreter directly in code
        p = sew.blobInterpreter.BlobInterpreter(
            [('p1', 'u8'), ('p2', 'i64'), ('p3', 'f64')]
        )

        # Select from the table
        self.d[self.tablename].select("*")
        result = self.d.fetchone()['data']

        # Interpret it
        interpreted = p.interpret(result)

        # Compare
        for k in self.data:
            self.assertEqual(interpreted[k], self.data[k])

    #%%
    def test_interpret_config(self):
        # Construct interpreter from config file
        p = sew.blobInterpreter.BlobInterpreter.fromConfig(
            os.path.join(os.path.dirname(__file__), "blobcfg.ini"), 
            "test")

        # Select from the table
        self.d[self.tablename].select("*")
        result = self.d.fetchone()['data']

        # Interpret it
        interpreted = p.interpret(result)

        # Compare
        for k in self.data:
            self.assertEqual(interpreted[k], self.data[k])

    #%%
    def test_interpret_dictionary(self):
        # Construct interpreter from a dictionary
        structure = {
            'p1': 'u8',
            'p2': 'i64',
            'p3': 'f64'
        }
        p = sew.blobInterpreter.BlobInterpreter.fromDictionary(structure)

        # Select from the table
        self.d[self.tablename].select("*")
        result = self.d.fetchone()['data']

        # Interpret it
        interpreted = p.interpret(result)

        # Compare
        for k in self.data:
            self.assertEqual(interpreted[k], self.data[k])

    #%%
    def test_interpret_sqlitestatement(self):
        # Construct interpreter from dictionary
        structure = {
            'p1': 'u8',
            'p2': 'i64',
            'p3': 'f64'
        }
        p = sew.blobInterpreter.BlobInterpreter.fromDictionary(structure)

        # Now generate statement fragments that correspond to the different components
        # For now, don't stringify them
        stmtfrags = p.generateSplitStatement("data")
        # print(stmtfrags)

        # Select from the table using the fragments
        stmt = "select %s from %s" % (",".join(stmtfrags), self.tablename)
        # print(stmt)
        self.d.execute(stmt)
        result = self.d.fetchone()

        # Check result
        for key in result.keys(): # Remember, sqlite3.Row default iterator is not the keys, so specify .keys()
            self.assertEqual(
                result[key], 
                self.data[key].tobytes()
        )
        
        # Now let's try the stringified version
        stmtfrags = p.generateSplitStatement("data", hexOutput=True)
        
        # Select again
        stmt = "select %s from %s" % (",".join(stmtfrags), self.tablename)
        # print(stmt)
        self.d.execute(stmt)
        result = self.d.fetchone()

        # Check result, failing
        for key, value in self.data.items():
            self.assertEqual(
                result[key], 
                p.hexifyBlob(value.tobytes())
            )

        

#%%
if __name__ == '__main__':
    unittest.main()


import numpy as np
import sys
import configparser


#%%
class BlobInterpreter:
    """
    Class to interpret a blob.

    Interpreting a blob involves a known structure, which specifies
    1) Descriptor (short string describing this field)
    2) Type (word length i.e number of bits, and the type to cast to)

    Internally, this is kept as a list of tuples.

    This can be used to automatically generate multiple columns when selecting from a blob.

    Dynamically sized sections are also supported.
    To specify a dynamically sized section, suffix the descriptor with "_dyn".

    There are 3 scenarios of dynamically sized sections:
    ------------------------------------------------------
    1) "Flex" arrays: only 1 of these are allowed per structure.
    These will attempt to occupy as much of the blob as possible.
    For example, a structure with
        a: u8
        b_dyn: u8
        c: u16

    will interpret a 10-byte blob as having a 7-byte array assigned to "b_dyn".
    ------------------------------------------------------
    2) "Externally fixed length" arrays: the length is specified in the descriptor as well,
    like "_3_dyn".

    For example, a structure with
        a: u8
        b_3_dyn: u8
        c: u16

    requires a 6-byte blob, where "b_3_dyn" is a 3-byte array
    from the 2nd byte to the 4th byte i.e. [1:4]
    ------------------------------------------------------
    3) "Internally fixed length" arrays: the length is specified in another field.
    These are denoted with "_?_dyn", and the associated field must have the same name,
    but suffixed with "_len".

    For example, a structure with
        a_len: u8
        a_?_dyn: u8

    will read "a_len" for the number of elements, then use that as the size of
    "a_?_dyn". This is commonly found in headers of packets.
    """

    # Define type dictionary
    STR_TO_TYPE = {
        'u8': np.uint8,
        'u16': np.uint16,
        'u32': np.uint32,
        'u64': np.uint64,
        'i8': np.int8,
        'i16': np.int16,
        'i32': np.int32,
        'i64': np.int64,
        'f32': np.float32,
        'f64': np.float64,
        'fc32': np.complex64,
        'fc64': np.complex128
    }

    STR_TO_SIZE = {
        'u8': 1,
        'u16': 2,
        'u32': 4,
        'u64': 8,
        'i8': 1,
        'i16': 2,
        'i32': 4,
        'i64': 8,
        'f32': 4,
        'f64': 8,
        'fc32': 8,
        'fc64': 16
    }

    STR_TO_CSTR = {
        'u8': '%hhu',
        'u16': '%hu',
        'u32': '%u',
        'u64': '%lu',
        'i8': '%hhd',
        'i16': '%hd',
        'i32': '%d',
        'i64': '%ld',
        'f32': '%f',
        'f64': '%f',
        # 'fc32': '%f', # These are a little complicated, let's not handle them for now
        # 'fc64': '%f
    }

    def __init__(self, structure: list=[]):
        """
        Creates a BlobInterpreter based on a specified structure.

        Parameters
        ----------
        structure : list
            Ordered list of tuples, where each tuple is of the form (descriptor, typestr).
            Type strings are specified in STR_TO_TYPE.
            Descriptors are fieldnames, usually used to identify the purpose of that section of data.
        """
        if not isinstance(structure, list):
            raise TypeError('Structure must be a list of tuples')
        self._structure = structure

    @classmethod
    def fromDictionary(cls, structure: dict):
        """
        Generates a BlobInterpreter from a dictionary.
        Dictionaries are expected to be ordered as of Python 3.7.
        """
        version = sys.version_info
        if version.major < 3 or (version.major == 3 and version.minor < 7):
            raise TypeError("Dictionary interpretation requires Python 3.7 or higher")
        return cls([(k, v) for k, v in structure.items()])
    
    @classmethod
    def fromConfig(cls, configfilepath: str, sectionname: str):
        """
        Generates a BlobInterpreter from a configuration file.
        This is loaded with 'configparser'; see https://docs.python.org/3/library/configparser.html.
        Config files are expected to be ordered as of Python 3.7.
        """
        version = sys.version_info
        if version.major < 3 or (version.major == 3 and version.minor < 7):
            raise TypeError("Configparser interpretation requires Python 3.7 or higher")

        cfg = configparser.ConfigParser()
        cfg.read(configfilepath)
        section = cfg[sectionname]
        return cls([(k, v) for k, v in section.items()])

    def appendField(self,  descriptor: str, type: str='u8'):
        """
        Appends a field to the structure.

        Parameters
        ----------
        descriptor : str
            Description of the field.
        type : str
            String that specifies the type to cast to.
            Defaults to 'u8'.
        """
        self._structure.append((descriptor, type))

    def interpret(self, blob: bytes) -> dict:
        """
        Interprets a blob and returns a dict of arrays according to the structure.

        Parameters
        ----------
        blob : bytes
            Bytes object, usually obtained from a BLOB column select.

        Returns
        -------
        output : dict
            Dictionary of arrays, according to the internal structure.
        """

        output = dict()
        ptr = 0
        for desc, typestr in self._structure:
            # Turn it into a numpy array
            interpreted = np.frombuffer(
                blob[ptr:ptr+self.STR_TO_SIZE[typestr]],
                dtype=self.STR_TO_TYPE[typestr]
            )

            ptr += self.STR_TO_SIZE[typestr]

            output[desc] = interpreted

        return output
    
    def generateSplitStatement(self, blobColumnName: str, hexOutput: bool=False):
        """
        Generates SQL statement fragments that correspond to 
        splitting a blob into multiple columns based on the structure.

        Parameters
        ----------
        blobColumnName : str
            The SQLite column name that contains the BLOB.
        hexOutput : bool
            Flag to determine whether to return as hex strings, useful for views.
            The default is False, which will return as raw BLOBs.
        """
        output = []
        ptr = 1 # Sqlite substr starts from 1
        for desc, typestr in self._structure:
            size = self.STR_TO_SIZE[typestr]
            fragment = f'substr({blobColumnName},{ptr},{size})'
            if hexOutput:
                fragment = f'hex({fragment})'
            fragment = f'{fragment} AS {desc}'
            output.append(fragment)
            ptr += size

        return output
    
    def hexifyBlob(self, blob: bytes) -> str:
        """
        Returns a hex string form of the blob, akin to what SQLite
        would produce with its hex() function.

        Parameters
        ----------
        blob : bytes
            Input bytes object. This may come from a slice of the
            SQLite BLOB column.

        Returns
        -------
        h : str
            Hex string formatted with %02X.
        """
        h = "".join(["%02X" % i for i in blob])

        return h



        

    


# -*- coding: utf-8 -*-
"""
Created on Mon Dec 19 16:41:17 2022

@author: seo
"""

import re

#%%
class FormatSpecifier:
    """
    Formats are generated as a dictionary of 'cols' and 'conds'
    which specify the columns and conditions respectively.

    Example:
    fmt = {
        'cols': [
            ['col1', 'integer'], # list of lists, with inner list begin column name and sqlite type in that order
            ['col2', 'REAL'] # the type upper/lower-case doesn't matter
        ],
        'conds': [
            "UNIQUE(col1, col2)" # this is just a list of strings, with each one specifying an extra condition
        ]
    }

    Foreign keys are also allowed; specify these as a list of lists, with each inner list specifying the column name
    and then the parent class/table name pair that it references.

    Example:
    foreign_key = [
        ["col_child", "parent_table(col_parent)"]
    ]

    becomes

    "FOREIGN KEY(col_child) REFERENCES parent_table(col_parent)"
    """


    # Constant statics
    sqliteTypes = {
        int: "INTEGER",
        float: "REAL",
        str: "TEXT",
        bytes: "BLOB"
    }
    keywordTypes = [
        "INTEGER", "INT", "REAL", "TEXT", "BLOB",
        "DOUBLE", "FLOAT", "NUMERIC"] # Non-exhaustive list of keyword types
    
    # Constructor
    def __init__(self, cols: list=[], conds: list=[], foreign_keys: list=[]):
        self.fmt = {'cols': cols, 'conds': conds, 'foreign_keys': foreign_keys}
        
    def __repr__(self):
        return str(self.fmt)
        
    def clear(self):
        self.fmt = {'cols': [], 'conds': [], 'foreign_keys': []}
        
    def _getColumnNames(self):
        return [i[0] for i in self.fmt['cols']]
        
    def addColumn(self, columnName: str, typehint: type):
        self.fmt['cols'].append([columnName, self.sqliteTypes[typehint]])
        
    def addUniques(self, uniqueColumns: list):
        if not all((i in self._getColumnNames() for i in uniqueColumns)):
            raise ValueError("Invalid column found.")
        self.fmt['conds'].append("UNIQUE(%s)" % (','.join(uniqueColumns)))

    def addForeignKey(self, childParentPair: list):
        """
        Appends a single foreign key to the list of foreign keys.
        Note, this does not check for existence of the parent table/column.

        Parameters
        ----------
        childParentPair : list
            A list of two strings, the child column name and the parent table/column name.
        """
        if childParentPair[0] not in self._getColumnNames():
            raise ValueError("Invalid child column found.")
        self.fmt['foreign_keys'].append(childParentPair)

        
    def generate(self):
        return self.fmt
    
    @classmethod
    def fromSql(cls, stmt: str):
        '''
        Factory method to create format specifier from a CREATE TABLE statement.

        Parameters
        ----------
        stmt : str
            The create table statement.
        '''
        # Pull out everything after tablename, remove parentheses
        fmtstr = re.search(r"\(.+\)", stmt.replace("\n","").replace("\r","")).group()[1:-1] # Greedy regex 
        # Remove any uniques
        uniques = re.finditer(r"UNIQUE\(.+?\)", fmtstr, flags=re.IGNORECASE) # Non-greedy regex
        conds = []
        for unique in uniques:
            fmtstr = fmtstr.replace(unique.group(), "") # Drop the substring
            conds.append(unique.group())
        
        # Remove any foreign keys
        foreignkeys = re.finditer(r"FOREIGN KEY(.+?) REFERENCES (.+?)\)", fmtstr, flags=re.IGNORECASE)
        foreign_keys = []
        for foreign in foreignkeys:
            fmtstr = fmtstr.replace(foreign.group(), "") # Drop the substring
            # Get the child column name by searching the first brackets
            childCol = re.search(r"\(.+?\)", foreign.group(), flags=re.IGNORECASE).group()[1:-1]
            # Get the parent table/column name by taking everything after REFERENCES
            parentColStart = re.search(r"REFERENCES ", foreign.group(), flags=re.IGNORECASE).span()[1]
            parentCol = foreign.group()[parentColStart:]
            foreign_keys.append([childCol, parentCol])

        # There are some problems with the old way of getting the columns
        # To be safe, we use another regex that extracts based on the expected types
        cols = re.finditer(
            r"(\w+)\s(%s)" % "|".join(cls.keywordTypes), fmtstr, flags=re.IGNORECASE
        )
        cols = [i.group().split() for i in cols]

        # Note, due to pythonic default arguments only evaluating at definition time,
        # we must take extra care to input all __init__ arguments here! Or else the default argument
        # will 'remember' the previous value.
        return cls(cols, conds, foreign_keys)

    @staticmethod
    def dictContainsColumn(fmt: dict, colname: str):
        '''Checks if a generated format dictionary contains a particular column.'''
        check = False
        for i, col in enumerate(fmt['cols']):
            if col[0] == colname:
                check = True
                break
        return check
    
    @staticmethod
    def getParents(fmt: dict):
        parents = dict()
        for keydesc in fmt['foreign_keys']:
            spl = keydesc[1].split("(")
            tablename = spl[0]
            columnname = spl[1][:-1]
            # For the weird cases where the same parent column
            # is pointed to by two child columns in the same table
            if (tablename, columnname) not in parents:
                parents[(tablename, columnname)] = []
            parents[(tablename, columnname)].append(keydesc[0]) # Map parent -> child column
        return parents
        

#%%
if __name__ == "__main__":
    fmtspec = FormatSpecifier()
    fmtspec.addColumn('col1', int)
    fmtspec.addColumn('col2', str)
    
    try:
        fmtspec.addUniques(['col2', 'col3'])
    except Exception as e:
        print(e)
        print("Should raise exception for invalid column.")
        
    fmtspec.addUniques(['col1', 'col2'])
    
    # Add more uniques
    fmtspec.addColumn('col3', float)
    fmtspec.addColumn('col4', float)
    fmtspec.addUniques(['col3', 'col4'])

    # Add some foreign keys
    fmtspec.addForeignKey(['col1', 'parent_table(parentcolA)'])
    fmtspec.addForeignKey(['col2', 'parent_table(parentcolB)'])

    print(fmtspec.generate())
    
    #%% Test fromSql
    from ._core import StatementGeneratorMixin
    
    stmt = StatementGeneratorMixin._makeCreateTableStatement(fmtspec.generate(), 'table1')
    print(stmt)
    genFmtspec = FormatSpecifier.fromSql(stmt)
    print("\n\n\n")

    print(genFmtspec.generate())
    # print(genFmtspec.generate()['foreign_keys'])
    # print(id(genFmtspec.fmt))
    # print(fmtspec.generate())
    # print(id(fmtspec.fmt))
    assert(genFmtspec.fmt == fmtspec.fmt)


import unittest
from ._helpers import *
import time
import numpy as np
import itertools

class TestBenchmarks(unittest.TestCase):
    def setUp(self):
        self.d = sew.Database(":memory:")
        self.fmtspec = sew.FormatSpecifier(
            [
                ["col1", "REAL"],
                ["col2", "REAL"]
            ]
        )

        self.fmtspeclong = sew.FormatSpecifier(
            [
                ["col1", "REAL"],
                ["col2", "REAL"],
                ["col3", "INTEGER"],
                ["col4", "TEXT"]
            ]
        )

        self.d.createTable(
            self.fmtspec.generate(),
            "benchmark"
        )
        self.d.createTable(
            self.fmtspeclong.generate(),
            "benchmarklong"
        )

        self.d.reloadTables()
        # print("Running tests.benchmarks")

    # def tearDown(self):
    #     print("Completed tests.benchmarks")

    # def test_benchmarks_10000(self):
    #     length = 10000
    #     t1 = time.time()
    #     self.d['benchmark'].insertMany(
    #         ((i, i+1) for i in range(length)),
    #         commitNow=True
    #     )
    #     t2 = time.time()
    #     print("%d generator inserts at %f/s." % (length, length/(t2-t1)))

    #     # Don't actually need to assert anything

    def test_benchmarks_1000000(self):
        length = 1000000
        t1 = time.time()
        self.d['benchmark'].insertMany(
            ((i, i+1) for i in range(length)),
            commitNow=True
        )
        t2 = time.time()
        print("%d generator inserts (2 cols) at %f/s." % (length, length/(t2-t1)))
        
        # Don't actually need to assert anything

    def test_benchmarks_arr1000000(self):
        length = 1000000
        data = np.random.rand(2, length)

        # This is signficantly slower, regardless of the method of iteration
        t1 = time.time()
        self.d['benchmark'].insertMany(
            np.nditer([data[0,:],data[1,:]], op_flags=[['readonly'],['readonly']]),
            # zip(data[0,:], data[1,:]),
            # ((data[0,i], data[1,i]) for i in range(length)), # All no significant change
            commitNow=True
        )
        t2 = time.time()
        print("%d array reference inserts (2 cols) at %f/s." % (length, length/(t2-t1)))

        # What if we transpose first? No difference..
        data = np.ascontiguousarray(data.T)
        t1 = time.time()
        
        self.d['benchmark'].insertMany(
            np.nditer([data[:,0], data[:,1]], op_flags=[['readonly'],['readonly']]),
            # data, # nditer offers no speedup
            commitNow=True
        )
        t2 = time.time()
        print("%d array (transposed) reference inserts (2 cols) at %f/s." % (length, length/(t2-t1)))

        # Compare this with inserting with insertOne
        t1 = time.time()
        for i in range(data.shape[0]):
            self.d['benchmark'].insertOne(
                data[i,0], data[i,1],
                commitNow=False
            )
        self.d.commit()
        t2 = time.time()
        print("%d array reference inserts (2 cols) performed using insertOne at %f/s." % (length, length/(t2-t1)))

        # Don't actually need to assert anything

    def test_benchmarkslong_1000000(self):
        length = 1000000
        t1 = time.time()
        self.d['benchmarklong'].insertMany(
            ((i, i+1, i+2, str(i)[0]) for i in range(length)),
            commitNow=True
        )
        t2 = time.time()
        print("%d generator inserts (4 cols) at %f/s." % (length, length/(t2-t1)))


if __name__ == "__main__":
    unittest.main()


# Import this for type hints with classes
from __future__ import annotations

class Condition:
    """
    Helper class to generate SQLite condition substrings.

    SQLite SELECT statements often have conditions. This class attempts to
    generate the condition substrings for any number of conditions, by chaining them together
    using pythonic operators.

    This works by always instantiating a Condition object, and then using standard comparison operators
    like ==, !=, >, <, >= and <=. At its heart, the class simply contains an extending string, and mutates this
    internally after every operator.

    Examples:

    Condition("col1") == '5'
    >> col1 = 5

    Condition("col1") > '5'
    >> col1 > 5

    You can also then chain multiple conditions together by doing it step by step:

    Example:
    c = Condition("col1") == '5'
    c & 'col2'
    c > '10'
    >> col1 = 5 AND col2 > 10

    Usually, the operators are easy to insert directly into the string yourself, so you can just use the
    &, | operators to splice multiple conditions together (this is the recommendation, as it clearly demarcates
    the different conditions and whether to AND or OR them together).:

    Example:
    c = Condition("col1 = 5")
    c & "col2 > 10"
    c | "col3 < 20"
    >> col1 = 5 AND col2 > 10 OR col3 < 20

    Note that due to the nature of Python's built-in operators, you can't place them all in one line unless you wrap each step in parentheses.
    
    Example:
    
    c = Condition("col1") == '5' & 'col2' > '10' # This does not work because Python will compare 'col2' > '10' separately which will give an error.
    c = (Condition("col1") == '5') & (Condition("col2") > '10') # This is a possible workaround, but is slightly verbose.
    c = ((Condition("col1") == '5') & "col2") > '10' # This is another workaround, which is less verbose but will require parentheses after every operator.

    c = Condition("col1 = 5") & Condition("col2 = 10") & Condition("col3 = 6") # No parentheses needed if objects completely wrap each substring, but again very verbose.
    c = (Condition("col1 = 5") & "col2 = 10") & "col3 = 6" # Some parentheses needed, but much less verbose.
    """
    def __init__(self, first: str):
        if not isinstance(first, str):
            raise TypeError("Condition must be a string.")
        self._cond = first

    def __str__(self) -> str:
        return self._cond
    
    def __repr__(self) -> str:
        return self._cond

    # There are a few SQLite conditions that don't really have an 'operator'
    def LIKE(self, other: Condition) -> Condition:
        # May be a string, in which case just attach it to the current condition
        if isinstance(other, str):
            self._cond = "%s LIKE %s" % (self._cond, other)
        
        # Otherwise mutate the current instance
        elif isinstance(other, Condition):
            self._cond = "%s LIKE %s" % (self._cond, other._cond)

        else:
            raise TypeError("Condition must be a string or Condition.")
        
        return self
    
    def IN(self, other: Condition) -> Condition:
        # May be a tuple or list of strings
        if isinstance(other, list) or isinstance(other, tuple):
            self._cond = "%s IN (%s)" % (self._cond, ",".join(other))
        
        # Otherwise mutate the current instance
        elif isinstance(other, Condition):
            self._cond = "%s IN (%s)" % (self._cond, ",".join(other._cond))

        else:
            raise TypeError("Condition must be a list/tuple or Condition.")
        
        return self
    
    # TODO: ALL, ANY, BETWEEN, EXISTS
    
    # Operator overloads
    def __and__(self, other: Condition) -> Condition:
        # May be a string, in which case just attach it to the current condition
        if isinstance(other, str):
            self._cond = "%s AND %s" % (self._cond, other)
        
        # Otherwise mutate the current instance
        elif isinstance(other, Condition):
            self._cond = "%s AND %s" % (self._cond, other._cond)

        else:
            raise TypeError("Condition must be a string or Condition.")
        
        return self

    def __or__(self, other: Condition) -> Condition:
        # May be a string, in which case just attach it to the current condition
        if isinstance(other, str):
            self._cond = "%s OR %s" % (self._cond, other)
        
        # Otherwise mutate the current instance
        elif isinstance(other, Condition):
            self._cond = "%s OR %s" % (self._cond, other._cond)

        else:
            raise TypeError("Condition must be a string or Condition.")
        
        return self
    
    def __eq__(self, other: Condition) -> Condition:
        # May be a string, in which case just attach it to the current condition
        if isinstance(other, str):
            self._cond = "%s = %s" % (self._cond, other)
        
        # Otherwise mutate the current instance
        elif isinstance(other, Condition):
            self._cond = "%s = %s" % (self._cond, other._cond)

        else:
            raise TypeError("Condition must be a string or Condition.")
        
        return self
    
    def __ne__(self, other: Condition) -> Condition:
        # May be a string, in which case just attach it to the current condition
        if isinstance(other, str):
            self._cond = "%s != %s" % (self._cond, other)
        
        # Otherwise mutate the current instance
        elif isinstance(other, Condition):
            self._cond = "%s != %s" % (self._cond, other._cond)

        else:
            raise TypeError("Condition must be a string or Condition.")
        
        return self
    
    def __gt__(self, other: Condition) -> Condition:
        # May be a string, in which case just attach it to the current condition
        if isinstance(other, str):
            self._cond = "%s > %s" % (self._cond, other)
        
        # Otherwise mutate the current instance
        elif isinstance(other, Condition):
            self._cond = "%s > %s" % (self._cond, other._cond)

        else:
            raise TypeError("Condition must be a string or Condition.")
        
        return self
    
    def __ge__(self, other: Condition) -> Condition:
        # May be a string, in which case just attach it to the current condition
        if isinstance(other, str):
            self._cond = "%s >= %s" % (self._cond, other)
        
        # Otherwise mutate the current instance
        elif isinstance(other, Condition):
            self._cond = "%s >= %s" % (self._cond, other._cond)

        else:
            raise TypeError("Condition must be a string or Condition.")
        
        return self
    
    def __lt__(self, other: Condition) -> Condition:
        # May be a string, in which case just attach it to the current condition
        if isinstance(other, str):
            self._cond = "%s < %s" % (self._cond, other)
        
        # Otherwise mutate the current instance
        elif isinstance(other, Condition):
            self._cond = "%s < %s" % (self._cond, other._cond)

        else:
            raise TypeError("Condition must be a string or Condition.")
        
        return self
    
    def __le__(self, other: Condition) -> Condition:
        # May be a string, in which case just attach it to the current condition
        if isinstance(other, str):
            self._cond = "%s <= %s" % (self._cond, other)
        
        # Otherwise mutate the current instance
        elif isinstance(other, Condition):
            self._cond = "%s <= %s" % (self._cond, other._cond)

        else:
            raise TypeError("Condition must be a string or Condition.")
        
        return self
    


# Load all imports using helper
from ._helpers import *

import unittest
import sqlite3 as sq
import numpy as np

#%%
class TestCorrectness(unittest.TestCase):
    def setUp(self):
        self.d = sew.Database(":memory:")
        self.fmtspec = sew.FormatSpecifier(
            [
                ["col1", "real"],
                ["col2", "real"],
                ["col3", "real"]
            ],
            ["UNIQUE(col1, col2)"]
        )
        self.d.createTable(
            self.fmtspec.generate(),
            "correctness"
        )
        # self.d.reloadTables()

    #%%
    def test_create_table_insert_drop(self):
        fmtspec = sew.FormatSpecifier()
        fmtspec.addColumn('c1', int)
        fmtspec.addColumn('c2', float)
        self.d.createTable(fmtspec.generate(),
                           'tbl')
        self.assertIn("tbl", self.d.tables)
        tbl = self.d['tbl']
        tbl.insertOne(
            10, 20.0
        )

        row = tbl[0]
        self.assertEqual(row['c1'], 10)
        self.assertEqual(row['c2'], 20.0)

        self.d.dropTable('tbl')
        self.assertNotIn("tbl", self.d.tables)


    #%%
    def test_formatSpecifier_getter(self):
        self.assertEqual(
            self.fmtspec.generate(),
            self.d["correctness"].formatSpecifier
        )

    #%%
    def test_insert_simple_and_delete(self):
        rows = [(10.0, 20.0, 30.0), (30.0,40.0,50.0)]
        self.d['correctness'].insertMany(
            rows, commitNow=True
        )
        # Check selected values
        self.d['correctness'].select("*")
        results = self.d.fetchall()
        for i, result in enumerate(results):
            for k in range(3):
                self.assertEqual(rows[i][k], result[k])

        # Delete one of the rows
        self.d["correctness"].delete(
            ["col2=20.0"], commitNow=True # This should remove one of the rows
        )
        # Check again
        self.d["correctness"].select("*")
        results = self.d.fetchall()
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0][0], rows[1][0])
        self.assertEqual(results[0][1], rows[1][1])
        self.assertEqual(results[0][2], rows[1][2])
        


    #%%
    def test_insert_generators(self):
        data1 = np.array([10.0, 30.0])
        data2 = np.array([30.0, 40.0])
        data3 = np.array([50.0, 60.0])
        self.d['correctness'].insertMany(
            ((data1[i], data2[i], data3[i]) for i in range(len(data1))),
            commitNow=True
        )
        # Check selected values
        self.d['correctness'].select("*")
        results = self.d.fetchall()
        for i, result in enumerate(results):
            self.assertEqual(data1[i], result[0])
            self.assertEqual(data2[i], result[1])
            self.assertEqual(data3[i], result[2])


    #%%
    def test_create_metadata(self):
        # First check that it throws if tablename or columns are wrong
        metaFmtspec = sew.FormatSpecifier(
            [
                ["data_tblname", "TEXT"],
                ["setting1", "INT"],
                ["setting2", "REAL"]
            ],
            []
        )

        with self.assertRaises(ValueError):
            self.d.createMetaTable(
                metaFmtspec.generate(),
                "wrong_tablename"
            )

        with self.assertRaises(ValueError):
            self.d.createMetaTable(
                self.fmtspec.generate(),
                "tbl_metadata"
            )

        # Now actually create a legitimate one
        self.d.createMetaTable(
            metaFmtspec.generate(),
            "tbl_metadata"
        )

        # Test creating a data table, make sure it throws if the associated meta table doesn't exist
        metadata = [5, 0.2]
        with self.assertRaises(ValueError):
            self.d.createDataTable(
                self.fmtspec.generate(),
                "mydata_table",
                metadata,
                "nonexistent_metadata"
            )

        # Now actually create the data table
        self.d.createDataTable(
            self.fmtspec.generate(),
            "mydata_table",
            metadata,
            "tbl_metadata"
        )

        # Check the list of data tables associated with this metadata table
        datatables = self.d['tbl_metadata'].getDataTables()
        self.assertListEqual(
            datatables,
            ["mydata_table"]
        )

        # Check that the data row exists inside metadata table
        metadataresult = self.d['tbl_metadata'].getMetadataFor(
            'mydata_table'
        )
        self.assertListEqual(
            metadata,
            [i for i in metadataresult]
        )

        # Also check that we can retrieve it from the data table
        metadataresultFromDataTable = self.d['mydata_table'].getMetadata()
        self.assertListEqual(
            metadata,
            [i for i in metadataresultFromDataTable]
        )

        # If this passes then by definition we should be able to see the classes were upgradecd correctly
        self.assertIsInstance(
            self.d['tbl_metadata'],
            sew.MetaTableProxy
        )
        self.assertIsInstance(
            self.d['mydata_table'],
            sew.DataTableProxy
        )

    #%%
    def test_redirect(self):
        self.assertEqual(
            self.d.cur.execute, self.d.execute,
            "execute redirect has failed")

    #%%
    def test_makeSelectStatement(self):
        tablename = "tablename"
        columnNames = ["col1", "col2"]
        conditions = ["col1 > ?", "col2 > ?"]
        orderBy = "col1 desc"
        stmt1 = self.d._makeSelectStatement(columnNames, tablename, encloseTableName=False)
        stmt2 = self.d._makeSelectStatement(columnNames, tablename, conditions, orderBy, encloseTableName=False)
        self.assertEqual(
            stmt1, "select col1,col2 from tablename",
            "select statement is incorrect"
        )
        self.assertEqual(
            stmt2, "select col1,col2 from tablename where col1 > ? and col2 > ? order by col1 desc",
            "select statement with conditions and ordering is incorrect"
        )

    #%%
    def test_uniqueness_throws(self):
        with self.assertRaises(sq.IntegrityError):
            self.d['correctness'].insertMany(
                [(0,1,2),(0,1,2)],
                orReplace=False
            )

    #%%
    def test_insert_requires_all_columns(self):
        with self.assertRaises(sq.ProgrammingError):
            self.d['correctness'].insertMany(
                ((i,) for i in range(2))
            )

        with self.assertRaises(sq.ProgrammingError):
            self.d['correctness'].insertOne(
                0.
            )

    #%%
    def test_insertOne_throws_if_enclosed(self):
        with self.assertRaises(TypeError):
            self.d['correctness'].insertOne((10.0, 20.0, 30.0))

    #%%
    def test_insert_namedcolumns(self):
        stmt = self.d['correctness'].insertOne(
            {
                'col1': 22.0,
                'col3': 44.0
            },
            commitNow=True
        )
        self.d['correctness'].select("*")
        # Ensure that the missing column comes back as None
        result = self.d.fetchone()
        self.assertEqual(result['col1'], 22.0)
        self.assertEqual(result['col2'], None)
        self.assertEqual(result['col3'], 44.0)

    #%%
    def test_makeCaseStatements(self):
        singlecase = self.d._makeCaseSingleConditionVariable(
            'col1', 
            [
                ['5', 'NULL'],
                ['10', '20']
            ],
            '-1'
        )
        checksinglecase = "CASE col1\n" \
            "WHEN 5 THEN NULL\n" \
            "WHEN 10 THEN 20\n" \
            "ELSE -1\n" \
            "END"
        self.assertEqual(singlecase, checksinglecase)

        multiplecase = self.d._makeCaseMultipleConditionVariables(
            [
                ['col1 < 5', 'NULL'],
                ['col2 > 10', '20']
            ],
            "-1"
        )
        checkmultiplecase = "CASE\n" \
            "WHEN col1 < 5 THEN NULL\n" \
            "WHEN col2 > 10 THEN 20\n" \
            "ELSE -1\n" \
            "END"
        self.assertEqual(multiplecase, checkmultiplecase)


    #%%
    def test_insertMany_namedcolumns(self):
        dictlist = [
            {'col1': 22.0, 'col3': 44.0},
            {'col1': 33.0, 'col3': 55.0}
        ]
        stmt = self.d['correctness'].insertManyNamedColumns(
            dictlist,
            commitNow=True
        )
        self.d['correctness'].select("*")
        # Ensure that the missing column comes back as None
        results = self.d.fetchall()
        for i, result in enumerate(results):
            self.assertEqual(result['col1'], dictlist[i]['col1'])
            self.assertEqual(result['col2'], None)
            self.assertEqual(result['col3'], dictlist[i]['col3'])

    #%%
    def test_createView(self):
        # Create a view with renames and amendments within select
        self.d['correctness'].createView(
            ["col1 as A", "(col2+10) as B"], # Warp the column value in the view too
        )
        # Reload to see the new view
        self.d.reloadTables()

        # Insert some data
        data = [(10.0, 20.0, 30.0), (30.0,40.0,50.0)]
        self.d['correctness'].insertMany(
            data
        )

        # Attempt to select from the new view
        self.d['correctness_view'].select("*")
        results = self.d.fetchall()
        for i, result in enumerate(results):
            self.assertIn("A", result.keys(), "A is a view column")
            self.assertIn("B", result.keys(), "B is a view column")
            self.assertEqual(
                result['A'], data[i][0]
            )
            self.assertEqual(
                result['B'], data[i][1] + 10 # Check the warped value
            )

    #%%
    def test_numeric_affinity(self):
        # Create a new table with NUMERIC affinity
        tblfmt = sew.FormatSpecifier(
            [
                ["col1", "NUMERIC"]
            ]
        )
        stmt = self.d.createTable(
            tblfmt.generate(), "tablenumeric")
        self.d.reloadTables()

        # Insert some data
        data = [(10,), (30.0,)]
        stmt = self.d["tablenumeric"].insertMany(
            data,
            commitNow=True
        )

        # Check results
        self.d["tablenumeric"].select("*")
        results = self.d.fetchall()
        for i, result in enumerate(results):
            self.assertEqual(result['col1'], data[i][0])

    #%%
    def test_basic_foreignkey(self):
        # Create a parent table
        parentfmt = sew.FormatSpecifier(
            [
                ["id", "INTEGER PRIMARY KEY"],
                ["val", "INTEGER"]
            ]
        )
        self.d.createTable(parentfmt.generate(), "parent")

        # Create a table with a foreign key
        fkfmt = sew.FormatSpecifier(
            [
                ["col1", "INTEGER"],
                ["col2", "INTEGER"]
            ],
            foreign_keys=[
                ["col2", "parent(id)"]
            ]
        )
        self.d.createTable(fkfmt.generate(), "child")

        self.d.reloadTables()

        # Insert into parent
        self.d["parent"].insertMany(
            [(1,2),(2,4)], commitNow=True
        )
        # Show that inserting into child fails if you don't use null
        with self.assertRaises(sq.IntegrityError):
            stmt = self.d["child"].insertOne(
                {"col1": 1, "col2": 10}, commitNow=True
            )
        # Show that inserting null is okay
        stmt = self.d["child"].insertOne(
            {"col1": 1, "col2": None}, commitNow=True
        )
        # Check that it returns as expected
        self.d["child"].select("*")
        result = self.d.fetchone()
        self.assertEqual(result['col1'], 1)
        self.assertEqual(result['col2'], None)

        # Insert one that actually references
        self.d["child"].insertOne(
            {"col1": 2, "col2": 2}, commitNow=True
        )

        # Then now show that removing from parent is not allowed
        with self.assertRaises(sq.IntegrityError):
            self.d["parent"].delete(
                ["id=2"], commitNow=True
            )
        # But removing the other one should be okay
        stmt = self.d["parent"].delete(
            ["id=1"], commitNow=True
        )
        self.d.execute(stmt)
        self.d.commit()

    #%%
    def test_foreignkey_parent_retrieval(self):
        # Create a parent table
        parentfmt = sew.FormatSpecifier(
            [
                ["id", "INTEGER PRIMARY KEY"],
                ["val", "INTEGER"]
            ]
        )
        self.d.createTable(parentfmt.generate(), "parent")

        # Create a table with a foreign key
        fkfmt = sew.FormatSpecifier(
            [
                ["col1", "INTEGER"],
                ["col2", "INTEGER"]
            ],
            foreign_keys=[
                ["col2", "parent(id)"]
            ]
        )
        self.d.createTable(fkfmt.generate(), "child")

        self.d.reloadTables()

        # Insert multiple rows into parent
        self.d["parent"].insertMany(
            [(1,2),(2,3)], commitNow=True
        )

        # Check parent results
        self.d["parent"].select("*")
        results = self.d.fetchall()
        # for result in results:
        #     print(dict(result))

        # Insert into child
        self.d["child"].insertOne(
            1, 1, commitNow=True
        )

        # Retrieve the child row
        self.d["child"].select("*")
        result = self.d.fetchone()
        # Retrieve an associated parent row
        self.d["child"].retrieveParentRow(result)
        parentResults = self.d.fetchall()
        for result in parentResults: # There should be only 1 result anyway
            self.assertEqual(result['id'], 1)
            self.assertEqual(result['val'], 2)

    #%%
    def test_multiple_foreignkeys(self):
         # Create a parent table
        parentfmt = sew.FormatSpecifier(
            [
                ["id", "INTEGER PRIMARY KEY"],
                ["val", "INTEGER"]
            ]
        )
        self.d.createTable(parentfmt.generate(), "parent")

        # Create another parent
        otherparentfmt = sew.FormatSpecifier(
            [
                ["id", "INTEGER PRIMARY KEY"],
                ["val", "INTEGER"]
            ]
        )
        self.d.createTable(otherparentfmt.generate(), "otherparent")

        # Create a table with two separate foreign keys
        fkfmt = sew.FormatSpecifier(
            [
                ["col1", "INTEGER"],
                ["col2", "INTEGER"],
                ["col3", "INTEGER"]
            ],
            foreign_keys=[
                ["col2", "parent(id)"],
                ["col3", "otherparent(id)"]
            ]
        )
        self.d.createTable(fkfmt.generate(), "child")

        self.d.reloadTables()

        # Insert into both parents
        self.d["parent"].insertMany(
            [(1,2),(2,3)], commitNow=True
        )
        self.d["otherparent"].insertMany(
            [(4,8),(5,10)], commitNow=True
        )

        # Then insert into the child
        self.d["child"].insertOne(
            1, 1, 4, commitNow=True
        )

        # Extract from the child
        self.d["child"].select("*")
        childResult = self.d.fetchone()

        # Extract the related parents
        self.d["child"].retrieveParentRow(childResult, "col2")
        parentResults = self.d.fetchall()
        for result in parentResults: # There should be only 1 result anyway
            self.assertEqual(result['id'], 1)
            self.assertEqual(result['val'], 2)

        self.d["child"].retrieveParentRow(childResult, "col3")
        otherparentResults = self.d.fetchall()
        for result in otherparentResults: # There should be only 1 result anyway
            self.assertEqual(result['id'], 4)
            self.assertEqual(result['val'], 8)

    #%%
    def test_foreignkey_relationships(self):
        # Create a bunch of parent tables
        parentfmt = sew.FormatSpecifier(
            [
                ["id", "INTEGER PRIMARY KEY"],
                ["val", "INTEGER"]
            ]
        )
        self.d.createTable(parentfmt.generate(), "parent1")
        self.d.createTable(parentfmt.generate(), "parent2")
        self.d.createTable(parentfmt.generate(), "parent3")

        # Create a bunch of child tables
        childfmt12 = sew.FormatSpecifier(
            [
                ["col1", "INTEGER"],
                ["col2", "INTEGER"],
                ["col3", "INTEGER"]
            ],
            foreign_keys=[
                ["col1", "parent1(id)"],
                ["col2", "parent2(id)"]
            ]
        )
        self.d.createTable(childfmt12.generate(), "child12")

        childfmt23 = sew.FormatSpecifier(
            [
                ["col1", "INTEGER"],
                ["col2", "INTEGER"],
                ["col3", "INTEGER"]
            ],
            foreign_keys=[
                ["col2", "parent2(id)"],
                ["col3", "parent3(id)"],
                ["col1", "parent3(id)"]
            ]
        )
        self.d.createTable(childfmt23.generate(), "child23")

        self.d.reloadTables()

        # Test the relationship dictionary
        families = self.d.relationships

        # Check for parent1
        self.assertTrue(
            ("child12", "col1") in families[("parent1", "id")]
        )

        # Check for parent2 (two children in diff tables to same parent)
        self.assertTrue(
            ("child23", "col2") in families[("parent2", "id")]
        )
        self.assertTrue(
            ("child12", "col2") in families[("parent2", "id")]
        )

        # Check for parent3 (two children in same table to same parent)
        self.assertTrue(
            ("child23", "col3") in families[("parent3", "id")]
        )
        self.assertTrue(
            ("child23", "col1") in families[("parent3", "id")]
        )

    #%%
    def test_table_bracket_access(self):
        self.d['correctness'].insertMany(
            [(i,i+1,i+2) for i in range(10)], commitNow=True
        )

        # Test a single row
        result = self.d['correctness'][3]
        self.assertEqual(result['col1'], 3)
        self.assertEqual(result['col2'], 4)
        self.assertEqual(result['col3'], 5)

        # Test slices
        results = self.d['correctness'][0:5]

        for i, result in enumerate(results):
            self.assertEqual(result['col1'], i)
            self.assertEqual(result['col2'], i+1)
            self.assertEqual(result['col3'], i+2)

        midresults = self.d['correctness'][5:8]
        for i, result in enumerate(midresults):
            self.assertEqual(result['col1'], i+5)
            self.assertEqual(result['col2'], i+6)
            self.assertEqual(result['col3'], i+7)

        # Test error if steps are provided
        with self.assertRaises(ValueError):
            self.d['correctness'][2:7:2]
    
    #%%
    def test_column_proxy_container(self):
        table = self.d['correctness']
        self.assertTrue(
            isinstance(table.columns, dict)
        )
        container = sew.ColumnProxyContainer(table.columns)
        # Check that we can access it like an attribute directly
        # print(container.col1)
        # print(table.columns['col1'])
        # ColumnProxy has an == override, but the object is identical so we use 'is' to compare
        self.assertIs(container.col1, table.columns['col1'])
        self.assertIs(container.col2, table.columns['col2'])
        self.assertIs(container.col3, table.columns['col3'])

    def test_context_manager(self):
        # Show that default sqlite3 doesn't close the database
        with sq.connect(":memory:") as sqdb:
            sqcur = sqdb.cursor()
        sqcur.execute("create table x(c1 INT)")

        # Then show that sew database does throw
        with sew.Database(":memory:") as db:
            cur = db.con.cursor()
        # Extracting a cursor manually
        with self.assertRaises(sq.ProgrammingError):
            cur.execute('create table x(c1 INT)')
        # Using our in-built executors
        with self.assertRaises(sq.ProgrammingError):
            db.execute('create table x(c1 INT)')



    #%% ==================================== PLUGINS ==================================== #
    def test_numpy_plugin(self):
        data_f64 = np.random.randn(100).astype(np.float64)
        data_f32 = np.random.randn(100).astype(np.float32)

        nd = sew.plugins.NumpyDatabase(":memory:")
        npfmtspec = sew.FormatSpecifier(
            [
                ["col1_f64", "real"],
                ["col2_f32", "real"]
            ]
        )

        nd.createTable(
            npfmtspec.generate(),
            "nptable"
        )
        nd.reloadTables()

        nd['nptable'].insertMany(
            data_f64, data_f32, commitNow=True
        )
        stmt = nd['nptable'].select("*")
        
        results = nd['nptable'].fetchAsNumpy()
        
        np.testing.assert_equal(data_f64, results['col1_f64'])
        np.testing.assert_equal(data_f32, results['col2_f32'])

    #%%
    def test_numpy_insertOne_throws(self):
        data_f64 = np.random.randn(1).astype(np.float64)
        data_f32 = np.random.randn(1).astype(np.float32)

        nd = sew.plugins.NumpyDatabase(":memory:")
        npfmtspec = sew.FormatSpecifier(
            [
                ["col1_f64", "real"],
                ["col2_f32", "real"]
            ]
        )

        nd.createTable(
            npfmtspec.generate(),
            "nptable"
        )
        nd.reloadTables()

        with self.assertRaises(NotImplementedError):
            nd['nptable'].insertOne(
                data_f64, data_f32, commitNow=True
            )


if __name__ == "__main__":
    unittest.main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 15 22:14:34 2023

@author: seoxubuntu
"""

from ._core import *
from .formatSpec import FormatSpecifier
import pandas as pd
import numpy as np

#%% Pandas plugins
class PandasCommonMethodMixin(CommonMethodMixin):
    def reloadTables(self):
        '''
        Loads and parses the details of all tables from sqlite_master.
        Pandas plugin version. Caches pandas-enabled tables internally.
        
        Returns
        -------
        results : 
            Sqlite results from fetchall(). This is usually used for debugging.
        '''
        stmt = self._makeSelectStatement(["name","sql"], "sqlite_master",
                                         conditions=["type='table'"])
        self.cur.execute(stmt)
        results = self.cur.fetchall()
        self._tables.clear()
        for result in results:
            self._tables[result[0]] = PandasTableProxy(self, result[0],
                                                       FormatSpecifier.fromSql(result[1]).generate())
            
        return results

class PandasTableProxy(TableProxy):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pdresults = None # Additional cache for pandas dataframe results
        
    @property
    def pdresults(self):
        '''Getter for pandas dataframe results.'''
        return self._pdresults
    
    def select(self,
               columnNames: list,
               conditions: list=None,
               orderBy: list=None):
        '''
        Performs a select on the current table.
        Pandas plugin version. Essentially forwards query to pd.read_sql().

        Parameters
        ----------
        columnNames : list
            List of columns to extract. A single column may be specified as a string.
            If all columns are desired, the string "*" may be specified.
            Examples:
                ["col1", "col3"]
                "*"
                "justThisColumn"
                
        conditions : list, optional
            The filter conditions placed after "where".
            A single condition may be specified as a string.
            The default is None, which will place no conditions.
            Examples:
                ["col1 < 10", "col2 = 5"]
                "justThisColumn >= 8"
                
        orderBy : list, optional
            The ordering conditions placed after "order by".
            A single condition may be specified as a string.
            The default is None, which will place no ordering on the results.
            Examples:
                ["col1 desc", "col2 asc"]
                "justThisColumn asc"

        Returns
        -------
        stmt : str
            The actual sqlite statement that was executed.
        '''
        
        stmt = self._makeSelectStatement(
            columnNames,
            self._tbl,
            conditions,
            orderBy
        )
        self._pdresults = pd.read_sql(stmt, self._parent.con) # Store into internals
        return stmt
    
    # TODO: complete insert variations for pandas, using the df.to_sql() command
        
class PandasDatabase(CommonRedirectMixin, PandasCommonMethodMixin, SqliteContainer):
    pass

#%% Numpy plugins %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class NumpyCommonMethodMixin(CommonMethodMixin):
    '''
    Design goals for Numpy plugin:
        1) Assume columns are marked with data type as a suffix.
        Example:
            "u64" -> np.uint64
            "f32" -> np.float32
            "fc128" -> np.complex128

        2) On selects, convert column into numpy arrays of the associated data type if specified.
        This means that the default behaviour is to return as columns as opposed to rows.

        3) On inserts, check for data type if column has suffix specified.
        
        4) Provide automatic dereferencing of array->column if inserts are done with multiple arrays.
    '''

    def reloadTables(self):
        '''
        Loads and parses the details of all tables from sqlite_master.
        Numpy plugin version.
        
        Returns
        -------
        results : 
            Sqlite results from fetchall(). This is usually used for debugging.
        '''
        stmt = self._makeSelectStatement(["name","sql"], "sqlite_master",
                                         conditions=["type='table'"])
        self.cur.execute(stmt)
        results = self.cur.fetchall()
        self._tables.clear()
        for result in results:
            self._tables[result[0]] = NumpyTableProxy(self, result[0],
                                                       FormatSpecifier.fromSql(result[1]).generate())
            
        return results

class NumpyTableProxy(TableProxy):
    numpyColumnSuffixes = {
        "u64": np.uint64,
        "u32": np.uint32,
        "u16": np.uint16,
        "u8": np.uint8,
        "i64": np.int64,
        "i32": np.int32,
        "i16": np.int16,
        "i8": np.int8,
        "f64": np.float64,
        "f32": np.float32,
        "f16": np.float16
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._npresults = None # Additional cache for numpy results
        
    def _getNumpyTypeFromSuffix(self, key: str):
        for suffix in self.numpyColumnSuffixes:
            if key[-len(suffix):] == suffix:
                return self.numpyColumnSuffixes[suffix]

        # Return None if no suffix
        return None

    def fetchAsNumpy(self, firstLength: int=100):
        '''
        Calls fetch and parses the data into numpy arrays for each column if suffixes are defined;
        if no appropriate suffix is found, it is returned as a row.

        Parameters
        ----------
        firstLength : int
            The first length of the arrays to preallocate.
            Due to the nature of sqlite, the number of rows returned is not known until 
            the fetch is complete. Hence, to minimise re-allocations, try to assign a
            number larger than the expected number of rows to this.
        '''

        # Require use of sqlite.Row
        if self._parent.con.row_factory is not sq.Row:
            raise ValueError("Cannot use fetchAsNumpy() unless sqlite3.Row is set as row_factory.")

        # We will return a dictionary of column names
        r = dict()
        # Iterate while fetching
        inited = False
        i = 0 # Our counter
        while True:
            # Fetch a row
            row = self._parent.cur.fetchone()
            if row is None:
                break # No more rows to fetch
            
            # On the first iteration fill the column names
            if not inited:
                for key in row.keys():
                    numpytype = self._getNumpyTypeFromSuffix(key)
                    if numpytype is not None:
                        # Instantiate a numpy array of appropriate type
                        r[key] = np.zeros(firstLength, dtype=numpytype)
                    else:
                        # Otherwise just make a list
                        r[key] = list()

                # Make sure we don't keep initing
                inited = True

            # Now iterate over the columns and write to respective arrays/lists
            for key in row.keys():
                # Check that our array has room left
                if i >= r[key].size:
                    # Otherwise replace with double the size
                    tmp = np.zeros(r[key].size * 2, dtype=r[key].dtype)
                    tmp[:r[key].size] = r[key][:]
                    r[key] = tmp
                
                # Write to our array
                try:
                    r[key][i] = row[key] # First try to just write directly, this should work if it returns a pythonic type
                except ValueError as e: # Otherwise it is a bytes-like object, try to parse it as a buffer to the correct dtype
                    r[key][i] = np.frombuffer(row[key], r[key].dtype)[0] # TODO: maybe we can perform it at the array level instead of the element level?


            # Increment counter
            i += 1

        return r

    def _numpyParseInserts(self, *args):
        '''
        Helper method to cast numpy arrays during inserts.
        This is mainly to ensure that sqlite can read the types,
        since it bugs out when the size is not "common".
        Example: 
            float32 will not be read in correctly, so it must be cast to float64.
        '''
        if hasattr(args[0], "__next__"): # Self-inputted generator is not allowed here
            raise TypeError("For numpy databases, pass in the individual arrays directly instead of your own generators.")

        length = args[0].size # They must all be the same length anyway
        _args = (
            [args[i][j] for i in range(len(args))]
            for j in range(length)
        )
        return _args

    def insertOne(self, *args, **kwargs):
        raise NotImplementedError("For numpy databases, you can only pass in numpy arrays via insertMany. Length one arrays are allowed there.")

    def insertMany(self, *args, orReplace: bool=False, commitNow: bool=False):
        '''
        Performs an insert statement for multiple rows of data.
        Note that this method assumes that a full insert is being performed
        i.e. all columns will have a value inserted.
        
        This version only supports numpy arrays, and assumes that individual columns are
        passed in as individual arrays in the *args parameters.
        Do not construct the generator object yourself, unlike the parent class.

        Parameters
        ----------
        *args : multiple numpy arrays
            Each argument should represent a column in the table.
            Example table:
                col1_f64 REAL, col2_f32 REAL
                ...
                arr_f64 = np.array([...], dtype=np.float64)
                arr_f32 = np.array([...], dtype=np.float32)
                insertMany(arr_f64, arr_f32)

        orReplace : bool, optional
            Overwrites the same data if True, otherwise a new row is created for every clash.
            The default is False.
            
        commitNow : bool, optional
            Calls commit on the database connection after the transaction if True. The default is False.

        Returns
        -------
        stmt : str
            The actual sqlite statement that was executed.

        '''

        # Make the generator
        _args = self._numpyParseInserts(*args)

        # Everything else is the same
        stmt = super().insertMany(_args, orReplace=orReplace, commitNow=commitNow)
        
        return stmt

class NumpyDatabase(CommonRedirectMixin, NumpyCommonMethodMixin, SqliteContainer):
    pass


#%%
if __name__ == "__main__":
    d = PandasDatabase(":memory:")
    
    fmt = {
        'cols': [
            ["col1", "integer"],
            ["col2", "DOUBLE"]
        ],
        'conds': [
            'UNIQUE(col1, col2)'    
        ]
    }
    
    # Test making tables
    d.createTable(fmt, 'table1')
    d.reloadTables()
    
    # Test inserting into table with dict-like access
    data = ((i, float(i+1)) for i in range(10)) # Generator expression
    print(d['table1'].insertMany(data))
    # Then check our results
    print(d['table1'].select("*"))
    # Retrieve by property and see a dataframe
    df = d['table1'].pdresults
    print(df)
    
    # What if i select only one column?
    print(d['table1'].select("col1"))
    print(d['table1'].pdresults) # Also works!


# -*- coding: utf-8 -*-
"""
Created on Mon Dec  5 17:23:06 2022

@author: icyveins7 
"""

import sqlite3 as sq
import re

from .formatSpec import FormatSpecifier

#%% Basic container, the most barebones
class SqliteContainer:
    def __init__(self, dbpath: str, row_factory: type=sq.Row, pragma_foreign_keys: bool=True):
        '''
        Instantiates an sqlite database container.

        Parameters
        ----------
        dbpath : str
            The database to connect to (either a filepath or ":memory:").
            See sqlite3.connect() for more information.
        row_factory : type, optional
            The row factory for the sqlite3 connection. The default is the in-built sqlite3.Row.
        pragma_foreign_keys : bool, optional
            Turns on PRAGMA FOREIGN_KEYS. The default is True.
        '''
        self.dbpath = dbpath
        self.con = sq.connect(dbpath)
        self.con.row_factory = row_factory
        self.cur = self.con.cursor()

        if pragma_foreign_keys:
            self.cur.execute("PRAGMA foreign_keys=ON")

    def __enter__(self):
        '''
        For use in a with statement.
        '''
        return self
    
    def __exit__(self, type, value, traceback):
        '''
        For use in a with statement.
        Closes the connection for you, unlike default sqlite3.Connection.
        '''
        self.con.close()
        
#%% Mixin to redirect common sqlite methods for brevity in code later
class CommonRedirectMixin:
    def __init__(self, *args, **kwargs):
        '''
        Provides several redirects to common methods, just to have shorter code.
        '''
        super().__init__(*args, **kwargs)
        
        self.close = self.con.close
        self.execute = self.cur.execute
        self.executemany = self.cur.executemany
        self.commit = self.con.commit
        self.fetchone = self.cur.fetchone
        self.fetchall = self.cur.fetchall
        self.fetchmany = self.cur.fetchmany
        
#%% Mixin that contains the helper methods for statement generation. Note that this builds off the standard format.
class StatementGeneratorMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    @staticmethod
    def _encloseTableName(tablename: str):
        return '"%s"' % tablename

    @staticmethod
    def _makeTableColumns(fmt: dict):
        return ', '.join([' '.join(i) for i in fmt['cols']])
    
    @staticmethod
    def _makeTableConditions(fmt: dict):
        return ', '.join(fmt['conds'])
    
    @staticmethod
    def _makeTableForeignKeys(fmt: dict):
        return ', '.join(
            ["FOREIGN KEY(%s) REFERENCES %s" % (i[0], i[1]) for i in fmt['foreign_keys']])
    
    @staticmethod
    def _makeCreateTableStatement(
        fmt: dict, tablename: str, ifNotExists: bool=False, encloseTableName: bool=True
    ):
        stmt = "create table%s %s(%s%s%s)" % (
            " if not exists" if ifNotExists else '',
            StatementGeneratorMixin._encloseTableName(tablename) if encloseTableName else tablename,
            StatementGeneratorMixin._makeTableColumns(fmt),
            ", %s" % (StatementGeneratorMixin._makeTableConditions(fmt)) if len(fmt['conds']) > 0 else '',
            ", %s" % (StatementGeneratorMixin._makeTableForeignKeys(fmt)) if fmt.get('foreign_keys') is not None and len(fmt['foreign_keys']) > 0 else ''
        )
        return stmt
    
    @staticmethod
    def _makeCreateViewStatemnt(
        selectStmt: str, viewtablename: str, ifNotExists: bool=False, encloseTableName: bool=True
    ):
        stmt = "create view%s %s as %s" % (
            " if not exists" if ifNotExists else '',
            StatementGeneratorMixin._encloseTableName(viewtablename) if encloseTableName else viewtablename,
            selectStmt
        )
        return stmt

    @staticmethod
    def _makeQuestionMarks(n: int): # fmt: dict):
        return ','.join(["?"] * n)
    
    @staticmethod
    def _makeNotNullConditionals(cols: dict):
        return ' and '.join(("%s is not null" % (i[0]) for i in cols))
    
    @staticmethod
    def _stitchConditions(conditions: list):
        conditionsList = [conditions] if isinstance(conditions, str) else conditions # Turn into a list if supplied as a single string
        conditionsStr = ' where ' + ' and '.join(conditionsList) if isinstance(conditionsList, list) else ''
        return conditionsStr
    
    @staticmethod
    def _makeCaseSingleConditionVariable(conditionVariable: str, whenthens: list, finalElse: str):
        """
        Used to generate a case statement for a single condition variable.
        By SQLite definitions, this can only test for multiple equality conditions, comparing the condition variable.

        This is usually paired with an "AS somenewcolumn".

        Example output:
            CASE colA
                WHEN 1 THEN 10
                WHEN 2 THEN 20
                ELSE -1
            END

        Parameters
        ----------
        conditionVariable : str
            Single string specifying the condition variable (does not need to be a column by itself).

        whenthens : list
            List of sublists of length 2: the first contains the WHEN substring
            and the second contains the THEN substring i.e.
            the first substring evaluates the equality of the condition variable,
            and the second substring will be the result if the equality evaluates to true.
            For this method, the first substring should only be a simple value for comparison.

        finalElse : str
            The final ELSE statement i.e. the value of the output column if no WHEN/THEN
            conditions evaluate to true.
        """
        # Stitch together the when/then sub statements
        whenthensStr = "\n".join(["WHEN %s THEN %s" % (i[0], i[1]) for i in whenthens])
        # Then combine with the final else statement and condition
        s = "CASE %s\n%s\nELSE %s\nEND" % (conditionVariable, whenthensStr, finalElse)

        return s
        
    @staticmethod
    def _makeCaseMultipleConditionVariables(whenthens: list, finalElse: str):
        """
        Used to generate a case statement for multiple condition variables.
        In this structure, every condition variable can be unique, and more general comparisons can
        be done like <, > etc.

        This is usually paired with an "AS somenewcolumn".

        Example output:
            CASE
                WHEN colA = 1 THEN 10
                WHEN colB > 2 THEN 20
                WHEN colC < 3 THEN 30
                ELSE -1
            END

        Parameters
        ----------
        whenthens : list
            List of sublists of length 2: the first contains the WHEN substring
            and the second contains the THEN substring i.e.
            the first substring evaluates the equality of the condition variable,
            and the second substring will be the result if the equality evaluates to true.
            For this method, the first substring can have more complicated expressions,
            which may involve not just equality comparisons and also involve multiple columns.

        finalElse : str
            The final ELSE statement i.e. the value of the output column if no WHEN/THEN
            conditions evaluate to true.
        """
        # Stitch together the when/then sub statements
        whenthensStr = "\n".join(["WHEN %s THEN %s" % (i[0], i[1]) for i in whenthens])
        # Then combine with the final else statement and condition
        s = "CASE\n%s\nELSE %s\nEND" % (whenthensStr, finalElse)

        return s
    
    @staticmethod
    def _makeSelectStatement(columnNames: list,
                             tablename: str,
                             conditions: list=None,
                             orderBy: list=None,
                             encloseTableName: bool=True):
        # Parse columns into comma separated string
        columns = ','.join(columnNames) if isinstance(columnNames, list) else columnNames
        # Parse conditions with additional where keyword
        conditions = StatementGeneratorMixin._stitchConditions(conditions)
        # Parse order by as comma separated string and pad the order by keywords
        orderBy = [orderBy] if isinstance(orderBy, str) else orderBy
        orderBy = ' order by %s' % (','.join(orderBy)) if isinstance(orderBy, list) else ''
        # Create the statement
        stmt = "select %s from %s%s%s" % (
                columns,
                StatementGeneratorMixin._encloseTableName(tablename) if encloseTableName else tablename,
                conditions,
                orderBy
            )
        return stmt
    
    @staticmethod
    def _makeInsertStatement(
        tablename: str, fmt: dict, orReplace: bool=False, encloseTableName: bool=True
    ):
        stmt = "insert%s into %s values(%s)" % (
                " or replace" if orReplace else '',
                StatementGeneratorMixin._encloseTableName(tablename) if encloseTableName else tablename,
                StatementGeneratorMixin._makeQuestionMarks(len(fmt['cols']))
            )
        return stmt
    
    @staticmethod
    def _makeInsertStatementWithNamedColumns(
        tablename: str, insertedColumns: list, orReplace: bool=False, encloseTableName: bool=True
    ):
        stmt = "insert%s into %s(%s) values(%s)" % (
            " or replace" if orReplace else '',
            StatementGeneratorMixin._encloseTableName(tablename) if encloseTableName else tablename,
            ','.join(insertedColumns),
            StatementGeneratorMixin._makeQuestionMarks(len(insertedColumns))
        )
        return stmt
    
    @staticmethod
    def _makeDropStatement(tablename: str):
        stmt = "drop table %s" % tablename
        return stmt
    
    @staticmethod
    def _makeDeleteStatement(
        tablename: str, conditions: list=None, encloseTableName: bool=True):
        stmt = "delete from %s%s" % (
            StatementGeneratorMixin._encloseTableName(tablename) if encloseTableName else tablename,
            StatementGeneratorMixin._stitchConditions(conditions)
        )
        return stmt
    
#%% We will not assume the CommonRedirectMixins here
class CommonMethodMixin(StatementGeneratorMixin):
    def __init__(self, *args, **kwargs):
        '''
        Includes common methods to create and drop tables, and provides
        dictionary-like access to all tables currently in the database.
        Also includes internal methods for statement generation.
        '''
        super().__init__(*args, **kwargs)
        
        self._tables = dict()
        self._relations = dict() # Establish in-memory parent->child mappings
        self.reloadTables()

    def _parseTable(
        self,
        table_name: str,
        table_sql: str,
        table_type: str
    ) -> dict:
        """
        Parses the parameters of a table and appropriately
        populates an internal dictionary with a particular
        table subclass object.

        This only handles TableProxy, MetaTableProxy and ViewProxy.
        For DataTableProxy, see _parseDataTable().

        Parameters
        ----------
        table_name : str
            Table name.
        table_sql : str
            SQLite statement used in generation of the table.
        table_type : str
            Table type, either 'table' or 'view'.

        Returns
        -------
        dataToMeta: dict
            An additional dictionary containing a lookup from
            DataTable: key -> MetaTable: value. Formed from querying
            the MetaTable object.
        """
        dataToMeta = dict()
        # Special case for view
        if table_type == 'view':
            self._tables[table_name] = ViewProxy(self, table_name)

        # Special cases for metadata tables
        elif table_name.endswith(MetaTableProxy.requiredTableSuffix):
            self._tables[table_name] = MetaTableProxy(
                self, table_name, 
                FormatSpecifier.fromSql(table_sql).generate())
            # We also retrieve all associated data table names
            assocDataTables = self._tables[table_name].getDataTables()
            for dt in assocDataTables:
                dataToMeta[dt] = table_name
        else:
            self._tables[table_name] = TableProxy(
                self, table_name, 
                FormatSpecifier.fromSql(table_sql).generate())

        return dataToMeta


    def _parseDataTable(
        self, 
        table_name: str,
        metatable_name: str
    ):
        """
        Parses a data table by re-instantiating it
        as a DataTable in the internal structure.

        Assumes it has already been created as a normal TableProxy.

        Parameters
        ----------
        table_name : str
            Name of the data table.
        metatable_name : str
            Name of its associated metadata table.
        """
        # Recreate the object as a DataTable instead
        self._tables[table_name] = DataTableProxy(
            self._tables[table_name]._parent,
            self._tables[table_name]._tbl,
            self._tables[table_name]._fmt,
            metatable_name
        )


    def _parseRelationship(
        self,
        tablename: str
    ):
        """
        Parses a table's parent-child relationships
        i.e. any foreign key relationships.

        Assumes the table is already present in the internal
        structure via ._parseTable().

        Parameters
        ----------
        tablename : str
            Target tablename.
        """
        if not isinstance(self._tables[tablename], ViewProxy):
            parents = FormatSpecifier.getParents(
                self._tables[tablename]._fmt)
            # Append to the relationship dict
            for parent, child_cols in parents.items():
                if parent not in self._relations.keys():
                    self._relations[parent] = list()
                for child_col in child_cols:
                    self._relations[parent].append((tablename,
                                                    child_col))


    def reloadTables(self):
        '''
        Loads and parses the details of all tables from sqlite_master.
        
        Returns
        -------
        results : 
            Sqlite results from fetchall(). This is usually used for debugging.
        '''
        stmt = self._makeSelectStatement(["name","sql","type"], "sqlite_master",
                                         conditions=["type='table' or type='view'"])
        self.cur.execute(stmt)
        results = self.cur.fetchall()
        self._tables.clear()

        dataToMeta = dict()
        for result in results:
            newDtm = self._parseTable(
                result['name'], result['sql'], result['type']
            )
            # Merge into dataToMeta
            dataToMeta.update(newDtm)

        # Iterate over all the tables to upgrade them to data tables if they exist
        for table in self._tables:
            if table in dataToMeta.keys():
                self._parseDataTable(
                    table,
                    dataToMeta[table]
                )

            # While doing so, check if it has foreign keys
            # But only if its not a view
            self._parseRelationship(table)
           
        return results
        
    def createTable(self, 
                    fmt: dict, 
                    tablename: str, 
                    ifNotExists: bool=False, 
                    encloseTableName: bool=True, 
                    commitNow: bool=False):
        '''
        Creates a new table.

        Parameters
        ----------
        fmt : dict
            Dictionary of column names/types and special conditions that characterises the table.
            The easiest way to generate this is to instantiate a FormatSpecifier object and then use
            generate() to create this.
        tablename : str
            The table name.
        ifNotExists : bool, optional
            Prevents creation if the table already exists. The default is False.
        encloseTableName : bool, optional
            Encloses the table name in quotes to allow for certain table names which may fail;
            for example, this is necessary if the table name starts with digits.
            The default is True.
        commitNow : bool, optional
            Calls commit on the database connection after the transaction if True. 
            The default is False.
        '''
        stmt = self._makeCreateTableStatement(fmt, tablename, ifNotExists, encloseTableName)
        self.cur.execute(stmt)
        if commitNow:
            self.con.commit()

        # Update the internal structure
        self._parseTable(tablename, stmt, 'table')
        return stmt

    def createMetaTable(self, 
                        fmt: dict, 
                        tablename: str, 
                        ifNotExists: bool=False, 
                        encloseTableName: bool=True, 
                        commitNow: bool=False):
        '''
        Creates a new meta table.

        Parameters
        ----------
        fmt : dict
            Dictionary of column names/types and special conditions that characterises the table.
            The easiest way to generate this is to instantiate a FormatSpecifier object and then use
            generate() to create this.
            Metadata tables are expected to have the first column designated as 'data_tblname TEXT'.
        tablename : str
            The table name.
        ifNotExists : bool, optional
            Prevents creation if the table already exists. The default is False.
        encloseTableName : bool, optional
            Encloses the table name in quotes to allow for certain table names which may fail;
            for example, this is necessary if the table name starts with digits.
            The default is True.
        commitNow : bool, optional
            Calls commit on the database connection after the transaction if True. 
            The default is False.
        '''

        # Check if the format and table names are valid
        if not tablename.endswith(MetaTableProxy.requiredTableSuffix):
            raise ValueError("Metadata table %s must end with %s" % (tablename, MetaTableProxy.requiredTableSuffix))

        if not FormatSpecifier.dictContainsColumn(fmt, MetaTableProxy.requiredColumn):
            raise ValueError("Metadata table %s must contain the column %s" % (tablename, MetaTableProxy.requiredColumn))

        # Otherwise, everything else is the same
        stmt = self._makeCreateTableStatement(fmt, tablename, ifNotExists, encloseTableName)
        self.cur.execute(stmt)
        if commitNow:
            self.con.commit()

        # Populate internal structure
        self._parseTable(tablename, stmt, 'table')

    def createDataTable(self, 
                        fmt: dict, 
                        tablename: str, 
                        metadata: list, 
                        metatablename: str,
                        metaOrReplace: bool=False,
                        ifNotExists: bool=False, 
                        encloseTableName: bool=True, 
                        commitNow: bool=False):
        '''
        Creates a new data table. This table will be intrinsically linked to a row in the associated metadata table.

        Parameters
        ----------
        fmt : dict
            Dictionary of column names/types and special conditions that characterises the table.
            The easiest way to generate this is to instantiate a FormatSpecifier object and then use
            generate() to create this.
        tablename : str
            The table name.
        metadata : list
            The metadata to insert into the metadata table associated with this data table.
        metatablename : str
            The metadata table name.
        metaOrReplace : bool, optional
            Whether to use orReplace when inserting into the metadata table.
            Useful when the metadata table has UNIQUE constraints.
        ifNotExists : bool, optional
            Prevents creation if the table already exists. The default is False.
        encloseTableName : bool, optional
            Encloses the table name in quotes to allow for certain table names which may fail;
            for example, this is necessary if the table name starts with digits.
            The default is True.
        commitNow : bool, optional
            Calls commit on the database connection after the transaction if True. 
            The default is False.
        '''

        # Check if the meta table exists
        if not metatablename in self._tables.keys():
            raise ValueError("Metadata table %s does not exist!" % metatablename)

        # Insert the associated metadata into the metadata table
        metastmt = self._makeInsertStatement(
            metatablename,
            self._tables[metatablename]._fmt,
            orReplace=metaOrReplace,
            encloseTableName=encloseTableName)
        metadata.insert(0, tablename) # The first argument is the name of the data table
        self.cur.execute(metastmt, metadata)

        # Otherwise, everything else is the same
        stmt = self._makeCreateTableStatement(fmt, tablename, ifNotExists, encloseTableName)
        self.cur.execute(stmt)
        if commitNow:
            self.con.commit()

        # Create the table internally
        self._parseTable(tablename, stmt, 'table')
        # Update it as a special DataTable
        self._parseDataTable(tablename, metatablename)



            
    def dropTable(self, tablename: str, commitNow: bool=False):
        '''
        Drops a table.

        Parameters
        ----------
        tablename : str
            The table name.
        commitNow : bool, optional
            Calls commit on the database connection after the transaction if True. 
            The default is False.
        '''
        self.cur.execute(self._makeDropStatement(tablename))
        if commitNow:
            self.con.commit()

        # Remove from internal structure
        self._tables.pop(tablename) # TODO: handle meta/data table complications?
    
    ### These are useful methods to direct calls to a table or query tables
    def __getitem__(self, tablename: str):
        return self._tables[tablename]
            
    @property
    def relationships(self):
        """
        Returns a dictionary of parent (key) to child (value) relationships.
        Each key is a 2-tuple of the form tablename,columnname.
        Values are returned as a list of 2-tuples, as multiple children may
        point to the same parent, even from within the same child table.

        Returns
        -------
        dict
            Parent to child relationship dictionary.
        """
        return self._relations

    @property
    def tablenames(self):
        '''
        List of names of current tables in the database.
        You may need to call reloadTables() if something is missing.
        '''
        return list(self._tables.keys())
    
    @property
    def tables(self):
        '''
        Dictionary of TableProxy objects that can be used individually to do table-specific actions.
        You may need to call reloadTables() if something is missing.
        '''
        return self._tables
    
#%% Akin to configparser, we create a class for tables
class TableProxy(StatementGeneratorMixin):
    def __init__(self, parent: SqliteContainer, tbl: str, fmt: dict):
        self._parent = parent # We redirect calls to the parent
        self._tbl = tbl # The tablename
        self._fmt = fmt
        self._cols = self._populateColumns()
        
    def _populateColumns(self):
        cols = dict()
        # typehints = 
        for col in self._fmt['cols']:
            colname = col[0]
            # Parse the type (note that we cannot determine the upper/lowercase)
            if re.match(r"int", col[1], flags=re.IGNORECASE): # All the versions have the substring 'int', so this works
                cols[colname] = ColumnProxy(colname, int)
            elif re.match(r"text", col[1], flags=re.IGNORECASE) or re.match(r"char", col[1], flags=re.IGNORECASE):
                cols[colname] = ColumnProxy(colname, str)
            elif re.match(r"real", col[1], flags=re.IGNORECASE) or re.match(r"double", col[1], flags=re.IGNORECASE) or re.match(r"float", col[1], flags=re.IGNORECASE):
                cols[colname] = ColumnProxy(colname, float)
            elif re.match(r"blob", col[1], flags=re.IGNORECASE):
                cols[colname] = ColumnProxy(colname, bytes)
            elif re.match(r"numeric", col[1], flags=re.IGNORECASE):
                cols[colname] = ColumnProxy(colname, (int, float))
            else:
                # cols[colname] = ColumnProxy(colname, object)
                raise NotImplementedError("Unknown parse for sql type %s" % col[1])
            
        return cols
    

    def __getitem__(self, i: slice):
        # For now, we don't have a built-in generator for limits and offsets
        # So we must build the statement ourselves
        if isinstance(i, int):
            self._parent.cur.execute(
                "SELECT * FROM %s LIMIT %d OFFSET %d" % (
                    self._tbl, 1, i
                )
            )
            results = self._parent.cur.fetchone()
        elif isinstance(i, slice):
            if i.step is not None and i.step != 1:
                raise ValueError("Cannot use a slice with a step")
            self._parent.cur.execute(
                "SELECT * FROM %s LIMIT %d OFFSET %d" % (
                    self._tbl, i.stop - i.start, i.start
                )
            )
            results = self._parent.cur.fetchall()
        return results

    @property
    def formatSpecifier(self):
        '''
        The format specifier for this table.
        Should effectively match the format specifier used during creation.
        '''
        return self._fmt
    
    @property
    def columns(self):
        '''
        Dictionary of ColumnProxy objects based on the table columns.
        '''
        return self._cols
    
    @property
    def columnNames(self):
        '''
        List of column names of the current table.
        '''
        return list(self._cols.keys())
 
    ### These are the actual user methods
    def select(self,
               columnNames: list,
               conditions: list=None,
               orderBy: list=None,
               encloseTableName: bool=True):
        '''
        Performs a select on the current table.

        Parameters
        ----------
        columnNames : list
            List of columns to extract. A single column may be specified as a string.
            If all columns are desired, the string "*" may be specified.
            Examples:
                ["col1", "col3"]
                "*"
                "justThisColumn"
                
        conditions : list, optional
            The filter conditions placed after "where".
            A single condition may be specified as a string.
            The default is None, which will place no conditions.
            Examples:
                ["col1 < 10", "col2 = 5"]
                "justThisColumn >= 8"
                
        orderBy : list, optional
            The ordering conditions placed after "order by".
            A single condition may be specified as a string.
            The default is None, which will place no ordering on the results.
            Examples:
                ["col1 desc", "col2 asc"]
                "justThisColumn asc"

        encloseTableName : bool, optional
            Encloses the table name in quotes to allow for certain table names which may fail;
            for example, this is necessary if the table name starts with digits.
            The default is True.

        Returns
        -------
        stmt : str
            The actual sqlite statement that was executed.
        '''
        
        stmt = self._makeSelectStatement(
            columnNames,
            self._tbl,
            conditions,
            orderBy,
            encloseTableName
        )
        self._parent.cur.execute(stmt)
        return stmt
    
    def delete(self, 
        conditions:list, 
        commitNow: bool=False, 
        encloseTableName: bool=True
    ):
        """
        Performs a delete statement on the current table.
        This does not have a LIMIT so all rows that fulfill the condition will be deleted.

        Parameters
        ----------
        conditions : list
            The filter conditions placed after "where".
            A single condition may be specified as a string.

        commitNow : bool, optional
            Calls commit on the database connection after the transaction if True. 
            The default is False.

        encloseTableName : bool, optional
            Encloses the table name in quotes to allow for certain table names which may fail;
            for example, this is necessary if the table name starts with digits.
            The default is True.
        """
        
        stmt = self._makeDeleteStatement(
            self._tbl, conditions, encloseTableName)
        self._parent.cur.execute(stmt)
        if commitNow:
            self._parent.con.commit()
        return stmt

    
    def insertOne(self, 
                  *args, 
                  orReplace: bool=False, 
                  commitNow: bool=False, 
                  encloseTableName: bool=True):
        '''
        Performs an insert statement for just one row of data.
        Note that this method assumes that a full insert is being performed
        i.e. all columns will have a value inserted.

        Parameters
        ----------
        *args : iterable or dict
            An iterable of the data for the row to be inserted.
            No need to place the arguments in a tuple,
            simply place them one after another before the keyword args.
            In this mode, all columns must have a value inserted 
            (see dict insertion below if you have missing values).
            Example:
                Two REAL columns
                insertOne(10.0, 20.0)

            Can also use a dictionary to used the named column format for insertion.
            Missing columns will have NULLs inserted as per sqlite's norm.
            Example:
                # Columns are ['col1','col2','col3']
                row = {
                    'col1': 11,
                    'col3': 22
                }
                insertOne(row)
                # After selection, the corresponding row will be
                # {'col1': 11, 'col2': None, 'col3': 22}
                
        orReplace : bool, optional
            Overwrites the same data if True, otherwise a new row is created.
            The default is False.
            
        commitNow : bool, optional
            Calls commit on the database connection after the transaction if True. 
            The default is False.

        encloseTableName : bool, optional
            Encloses the table name in quotes to allow for certain table names which may fail;
            for example, this is necessary if the table name starts with digits.
            The default is True.

        Returns
        -------
        stmt : str
            The actual sqlite statement that was executed.
        '''
        if isinstance(args[0], tuple) or isinstance(args[0], list):
            raise TypeError("Do not enclose the arguments in a list/tuple yourself!")
        
        if isinstance(args[0], dict):
            keys = list(args[0].keys())
            stmt = self._makeInsertStatementWithNamedColumns(
                self._tbl, keys, orReplace, encloseTableName
            )
            self._parent.cur.execute(stmt, [args[0][k] for k in keys])
    
        else:
            stmt = self._makeInsertStatement(
                self._tbl, self._fmt, orReplace, encloseTableName
            )
            self._parent.cur.execute(stmt, (args))

        if commitNow:
            self._parent.con.commit()
        return stmt
        
    def insertMany(self, 
                   rows: list, 
                   orReplace: bool=False, 
                   commitNow: bool=False, 
                   encloseTableName: bool=True):
        '''
        Performs an insert statement for multiple rows of data.
        Note that this method assumes that a full insert is being performed
        i.e. all columns will have a value inserted.

        Parameters
        ----------
        rows : iterable or generator expression
            An iterable or generator expression of the data of multiple rows. 
            It is assumed implicitly that every column has a value inserted i.e.
            no missing columns. 
            
            Dictionary mode insertion like in insertOne() is not supported as the
            statement would mutate on every row. See insertManyNamedColumns() for such use-cases.
            See sqlite3.executemany() for more information.

            Example with list of tuples/lists:
                Two REAL columns
                insertMany([(10.0, 20.0),(30.0, 40.0)])
            Example with generator:
                data1 = np.array([...])
                data2 = np.array([...])
                insertMany(
                    ((data1[i], data2[i]) for i in range(data1.size))
                )
            
        orReplace : bool, optional
            Overwrites the same data if True, otherwise a new row is created for every clash.
            The default is False.
            
        commitNow : bool, optional
            Calls commit on the database connection after the transaction if True. 
            The default is False.

        encloseTableName : bool, optional
            Encloses the table name in quotes to allow for certain table names which may fail;
            for example, this is necessary if the table name starts with digits.
            The default is True.

        Returns
        -------
        stmt : str
            The actual sqlite statement that was executed.
        '''
        stmt = self._makeInsertStatement(
            self._tbl, self._fmt, orReplace, encloseTableName
        )

        self._parent.cur.executemany(stmt, rows)

        if commitNow:
            self._parent.con.commit()
        return stmt
    
    def insertManyNamedColumns(self, 
                               dictlist: list, 
                               orReplace: bool=False, 
                               commitNow: bool=False,
                               encloseTableName: bool=True):
        '''
        Performs an insert statement for multiple rows of data.
        This method assumes that every row inserts the same set of named columns.

        Parameters
        ----------
        dictlist : list of dictionaries
            List of dictionaries where the keys are the columns being inserted.
            Each row corresponds to a dictionary. Missing columns will have NULLs inserted as per sqlite's norm.
            Example:
                # Rows are ['A','B','C']
                dictlist = [
                    {'A': 1, 'B': 2},
                    {'A': 4, 'B': 5}
                ]
                insertManyNamedColumns(dictlist)
            
        orReplace : bool, optional
            Overwrites the same data if True, otherwise a new row is created for every clash.
            The default is False.
            
        commitNow : bool, optional
            Calls commit on the database connection after the transaction if True.
            The default is False.

        encloseTableName : bool, optional
            Encloses the table name in quotes to allow for certain table names which may fail;
            for example, this is necessary if the table name starts with digits.
            The default is True.

        Returns
        -------
        stmt : str
            The actual sqlite statement that was executed.
        '''
        keys = list(dictlist[0].keys())
        stmt = self._makeInsertStatementWithNamedColumns(
            self._tbl, keys, orReplace, encloseTableName
        )
        # Create a generator for the list of dictionaries
        g = (
            [dictlist[i][k] for k in keys]
            for i in range(len(dictlist))
        )
        self._parent.cur.executemany(stmt, g)

        if commitNow:
            self._parent.con.commit()

        return stmt
    
    def createView(
        self,
        columnNames: list,
        conditions: list=None,
        orderBy: list=None,
        viewtbl_name: str=None,
        ifNotExists: bool=False,
        encloseTableName: bool=True,
        commitNow: bool=False
    ):
        '''
        Creates a view based on the current table.
        A view is essentially a pre-defined select statement, and is useful
        for looking at data without writing to an actual table, or pre-pending
        large select statements before other select statements.

        Parameters
        ----------
        columnNames : list
            Columns to extract as part of the select. See select().
        conditions : list, optional
            Conditions of the select. See select(). By default None.
        orderBy : list, optional
            Ordering of the select. See select(). By default None.
        viewtbl_name : str, optional
            The view's name, akin to a table name. Defaults to the current table with '_view' appended.
        ifNotExists : bool, optional
            Prevents creation if the view already exists. The default is False.
        encloseTableName : bool, optional
            Encloses the view name in quotes to allow for certain view names which may fail;
            for example, this is necessary if the view name starts with digits.
            The default is True.
        commitNow : bool, optional
            Calls commit on the database connection after the transaction if True.
            The default is False.
        
        Returns
        -------
        stmt : str
            The executed create view statement.
        '''
        # Default view name simply appends _view
        if viewtbl_name is None:
            viewtbl_name = self._tbl + "_view"

        # Generate select statement
        selectStmt = self._makeSelectStatement(
            columnNames,
            self._tbl,
            conditions,
            orderBy,
            encloseTableName
        )

        # Generate create view statement and execute
        stmt = self._makeCreateViewStatemnt(
            selectStmt,
            viewtbl_name,
            ifNotExists,
            encloseTableName
        )
        self._parent.cur.execute(stmt)
        if commitNow:
            self._parent.con.commit()

        return stmt
    
    ### Foreign-key specific methods
    def retrieveParentRow(self, row: sq.Row, foreignKey: str=None):
        """
        Performs a select on the associated parent table and row specified by the foreign key
        in the current child table. You are expected to call the fetch() flavours yourself afterwards.

        This is equivalent to performing

        "SELECT * FROM PARENT WHERE PARENT_KEY=SOME_VALUE"

        Parameters
        ----------
        row : sq.Row
            A row result returned from a select query on this current (child) table.
        foreignKey : str, optional
            The foreign key (column name) whose parent you seek, by default None,
            which just uses the first foreign key in the schema.

        Raises
        ------
        KeyError
            If foreign key is specified and does not exist.
        """
        # If no foreign key specified, assume the first foreign key in the schema
        if foreignKey is None:
            fk = self._fmt['foreign_keys'][0]
        else:
            fk = None
            for tfk in self._fmt['foreign_keys']:
                if tfk[0] == foreignKey:
                    fk = tfk
                    break
            if fk is None:
                raise KeyError(
                    f"The foreign key {foreignKey} does not exist in the table {self._tbl}"
                )
            
        # Extract the child column
        child_col = fk[0]

        # Extract the parent
        parent = fk[1]
        parentspl = parent.split("(")
        parent_table = parentspl[0]
        parent_col = parentspl[1][:-1]

        # Now perform a select on the parent table
        stmt = self._parent[parent_table].select("*", ["%s=%s" % (parent_col, str(row[child_col]))])
        return stmt

        
#%%
class ViewProxy(TableProxy):
    def __init__(self, parent: SqliteContainer, tbl: str):
        super().__init__(parent, tbl, None)

    def _populateColumns(self):
        return None
    
    def __getitem__(self, col: str):
        raise NotImplementedError("Invalid for view.")
    
    @property
    def columns(self):
        '''
        Dictionary of ColumnProxy objects based on the table columns.
        Not implemented for 
        '''
        raise NotImplementedError("Invalid for view.")
    
    @property
    def columnNames(self):
        '''
        List of column names of the current table.
        '''
        raise NotImplementedError("Invalid for view.")
    

#%% We have a special subclass for tables that are treated as metadata for other tables
# These tables contain a data_tblname column, and then all other columns are treated as metadata for it.
# This is especially useful if a table has a bunch of constant columns, but across tables these columns may have different values
# i.e. something like a table of processing results for a particular run, but each table used a different config file.
class MetaTableProxy(TableProxy):
    requiredColumn = 'data_tblname'
    requiredTableSuffix = "_metadata"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Metadata tables MUST end with "_metadata"
        if not self._tbl.endswith(self.requiredTableSuffix): # We throw if either created wrongly by user or accidentally internally
            raise ValueError("Metadata tables must end with '_metadata' but %s did not!" % self._tbl)

        # Format must contain 'data_tblname', and all other columns are treated as the actual metadata
        if not FormatSpecifier.dictContainsColumn(self._fmt, self.requiredColumn):
            raise ValueError("Format must contain 'data_tblname' as the primary key, but %s did not!" % self._fmt)

    def getMetadataFor(self, data_tblname: str):
        '''
        Returns the metadata for a particular data_tblname.

        Parameters
        ----------
        data_tblname : str
            The name of the data table to get the metadata for.

        Returns
        -------
        metadata : sqlite3.Row
            The metadata for the data_tblname.
        '''
        stmt = self._makeSelectStatement("*", self._tbl, ["data_tblname='%s'" % data_tblname])
        self._parent.cur.execute(stmt)
        metadata = self._parent.cur.fetchone()
        if metadata is None:
            raise ValueError("No metadata found for %s!" % data_tblname)
        return metadata

    def getDataTables(self):
        '''
        Returns a list of all the data tables associated to this metadata table.

        Returns
        -------
        data_tblnames : list
            A list of all the data tables in the database.
        '''
        stmt = self._makeSelectStatement("data_tblname", self._tbl)
        self._parent.cur.execute(stmt)
        data_tblnames = [row[0] for row in self._parent.cur.fetchall()]
        return data_tblnames

#%% Data tables act exactly like any other table, but keep track of their metadatatable internally
class DataTableProxy(TableProxy):
    def __init__(self, parent: SqliteContainer, tbl: str, fmt: dict, metadatatable: str=None):
        super().__init__(parent, tbl, fmt)
        self._metadatatable = metadatatable # Keeps track of metadatatable

    def setMetadataTable(self, metadatatable: str):
        '''Sets the metadata table name for this data table.'''
        self._metadatatable = metadatatable

    @property
    def metadataTablename(self):
        '''Returns the metadata tablename.'''
        return self._metadatatable

    def getMetadata(self):
        '''
        Returns the metadata for this data table by accessing the associated
        metadata table and extracting the relevant row.

        Returns
        -------
        metadata : sqlite3.Row
            Sqlite row result for the metadata.
            This will usually contain the current table's name as the first column.
        '''
        # We access the metadata table through the parent container
        return self._parent[self._metadatatable].getMetadataFor(self._tbl)
    
        

#%% And also a class for columns
### TODO: Intention for this is to build it into a way to automatically generate conditions in select statements..
class ColumnProxy:
    def __init__(self, name: str, typehint: type):
        self.name = name
        self.typehint = typehint
        
    def _requireType(self, x):
        if not isinstance(x, self.typehint):
            raise TypeError("Compared value must be of type %s" % str(self.typehint))
        
    def __lt__(self, x):
        self._requireType(x)
        return "%s < %s" % (self.name, str(x))

    def __le__(self, x):
        self._requireType(x)
        return "%s <= %s" % (self.name, str(x))
    
    def __gt__(self, x):
        self._requireType(x)
        return "%s > %s" % (self.name, str(x))

    def __ge__(self, x):
        self._requireType(x)
        return "%s >= %s" % (self.name, str(x))
    
    def __eq__(self, x):
        self._requireType(x)
        return "%s = %s" % (self.name, str(x))
    
    def __ne__(self, x):
        self._requireType(x)
        return "%s != %s" % (self.name, str(x))


class ColumnProxyContainer:
    def __init__(self, cols: dict[ColumnProxy]):
        self._cols = cols
        # Write a setattr for each column name
        for colProxy in self._cols:
            setattr(self, colProxy, self._cols[colProxy])



#%% Inherited class of all the above
class Database(CommonRedirectMixin, CommonMethodMixin, SqliteContainer):
    def __init__(self, dbpath: str, row_factory: type=sq.Row, pragma_foreign_keys: bool=True):
        '''
        Instantiates an sqlite database container with all extra functionality included.
        This enables:
            CommonRedirectMixin
                Provides several redirects to common methods, just to have shorter code.
            CommonMethodMixin
                Includes common methods to create and drop tables, and provides
                dictionary-like access to all tables currently in the database.
                Also includes internal methods for statement generation.
            
        Parameters
        ----------
        dbpath : str
            The database to connect to (either a filepath or ":memory:").
            See sqlite3.connect() for more information.
        row_factory : type, optional
            The row factory for the sqlite3 connection. The default is the in-built sqlite3.Row.
        pragma_foreign_keys : bool, optional
            Turns on PRAGMA FOREIGN_KEYS. The default is True.
        '''
        super().__init__(dbpath, row_factory, pragma_foreign_keys)

    
#%%
if __name__ == "__main__":
    d = Database(":memory:")
    tablename = "tablename"
    columnNames = ["col1", "col2"]
    conditions = ["col1 > ?", "col2 > ?"]
    orderBy = "col1 desc"
    
    print(d._makeSelectStatement(columnNames, tablename))
    print(d._makeSelectStatement(columnNames, tablename, conditions, orderBy))
    
    fmt = {
        'cols': [
            ["col1", "integer"],
            ["col2", "DOUBLE"]
        ],
        'conds': [
            'UNIQUE(col1, col2)'    
        ]
    }
    
    # Test making tables
    d.createTable(fmt, 'table1')
    
    # Test with no conditions
    fmt['conds'] = []
    d.createTable(fmt, 'table2')
    # Reload tables to populate internal dict then query
    d.reloadTables()
    print(d.tablenames)
    
    # Test delete statements
    print(d._makeDeleteStatement('table1'))
    print(d._makeDeleteStatement('table1', ['col1 < 10']))
    
    # Test dropping a table
    d.dropTable('table2', True)
    d.reloadTables()
    print(d.tablenames)
    
    # Test inserting into table with dict-like access
    data = ((i, float(i+1)) for i in range(10)) # Generator expression
    print(d['table1'].insertMany(data))
    # Then check our results
    print(d['table1'].select("*"))
    results = d.fetchall()
    for result in results:
        print(result[0], result[1])
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 17 19:33:25 2023

@author: seoxubuntu

Run this from the outer sew folder with 'python -m tests'
"""

import unittest

print("Begin tests")

# Import correctness unittests
print("Running correctness unittests")
from .correctness import *

# Import blob unittests
print("Running blob unittests")
from .blobs import *

# Run everything
unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 15 22:52:11 2023

@author: seoxubuntu
"""

from setuptools import setup, find_packages
setup(name="sew",
      version="0.0.4",
      packages=find_packages("."))

